#!/usr/bin/env python3
import os
import csv
import sys
import random
import hashlib
import platform
from time import time, strftime, gmtime
from datetime import datetime
from argparse import ArgumentParser, RawTextHelpFormatter, FileType
from functools import partial

from blessed import Terminal


filename = os.path.basename(sys.argv[0])
doc = f"""example:
  {filename} -i test.txt -s -d 60
  echo 'The typing seems really strong today.' | {filename} -d 3.5
  {filename} < test.txt

shortcuts:
  ^c / ctrl+c           end the test and get results now
  ^[ / ctrl+[ / esc     end the test and get results now
  ^h / ctrl+h / bksp    delete a character
  ^r / ctrl+r / tab     restart the same test
  ^s / ctrl+s           restart the test with words reshuffled
  ^w / ctrl+w           delete a word
  ^u / ctrl+u           delete a word
"""


def run():
    """Parse command arguments and run main"""
    main(**parse_args())


def main(
    duration,
    input,
    rows,
    shuffle,
    help,
    output_directory,
    hash,
):
    """Reads test words from `input` delimited by whitespace characters.
    Listens to standard input forming a typed word every time a
    whitespace is pressed. Compares typed words with test words. Updates the
    screen calling `draw` every time a key is pressed or every 0.1 seconds,
    whichever comes first.

    The test ends when `duration` time has passed or all words have been typed.
    Upon exiting, test results are printed and stored in `test_results_file`.
    """
    if input.isatty():  # no test words provided, fallback to a default test
        basedir = os.path.dirname(__file__)
        input = open(basedir + "/tests/common_300", "r")
        if duration is None:
            duration = 60
            shuffle = True

    if duration is None:
        duration = float("inf")

    test = input.read()
    words = test.split()

    if hash is None:
        hash = hashlib.sha1(test.encode("utf-8")).hexdigest()

    if shuffle:
        random.shuffle(words)

    if not sys.__stdin__.isatty():  # force stdin from user
        if platform.system() == "Windows":
            sys.__stdin__ = open("con:", "r")  # NOT TESTED
        else:
            sys.__stdin__ = open("/dev/tty", "r")

    term = Terminal()

    color_normal = term.normal
    color_correct = term.color_rgb(0, 230, 0)
    color_wrong = term.color_rgb(230, 0, 0)

    correct_chars = total_chars = -1
    wpm = 0  # words per minute
    time_passed = actual_duration = start = 0
    word_i = 0
    text = ""  # text the user writes
    colors = [color_normal] * len(words)

    char_times = []
    restart_count = 0

    with term.raw(), term.cbreak(), term.fullscreen(), term.hidden_cursor():
        while word_i < len(words) and (not start or time() - start < duration):
            word = words[word_i]

            if word == text:
                color = color_correct
            elif word.startswith(text):
                color = color_normal
            else:
                color = color_wrong

            colors[word_i] = color + term.reverse

            draw(term, rows, words, colors, word_i, text, wpm, time_passed)

            char = term.inkey(timeout=0.1, esc_delay=0)
            char_time = time()
            time_passed = char_time - start if start else 0

            if not char:
                continue

            if not start:
                start = time()

            if char == "\x03" or char == "\x1b":  # ctrl-c or ctrl-[ or esc
                # stop the test
                break

            elif char == "\x08" or char == "\x7f":  # ctrl-h or bksp
                # delete last character
                text = text[:-1]

            elif char in ("\x12", "\x13", "\t"):  # ctrl-r or ctrl-s or tab
                # restart test
                restart_count += 1
                correct_chars = total_chars = -1
                wpm = 0
                time_passed = actual_duration = start = 0
                word_i = 0
                text = ""
                colors = [color_normal] * len(words)
                char_times = []
                if char == "\x13":  # ctrl-s
                    random.shuffle(words)

            elif char == "\x15" or char == "\x17":  # ctrl-u or ctrl-w
                # clear user input
                text = ""

            elif char.isspace() and text:  # word is submitted
                if word_i + 1 < len(words):  # if not last space
                    # count the space character as correct
                    total_chars += 1
                    correct_chars += 1

                if text == word:
                    correct_chars += len(word)
                    colors[word_i] = color_correct
                else:
                    colors[word_i] = color_wrong

                actual_duration = time_passed
                wpm = min(int(correct_chars * 12 / actual_duration), 999)

                text = ""
                word_i += 1
                char_times.append((char, char_time))

            elif not char.isspace():
                # append the character to user input
                total_chars += 1
                text += char
                if word_i + 1 >= len(words) and words[-1] == text:  # last word
                    # end test without needing to submit a space
                    term.ungetch(" ")
                char_times.append((char, char_time))

    # remove excess user input
    total_chars -= len(text)

    if actual_duration <= 0 or total_chars <= 0:  # test is invalid
        return

    # calculate results and write them to output files

    accuracy = 100 * correct_chars // total_chars
    timestamp = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    print(f"accuracy: {accuracy}%")
    print(f"speed:    {wpm}wpm")
    print(f"duration: {actual_duration:.2f}s")
    print(f"restarts: {restart_count}")
    print(
        f"keystrokes: {correct_chars} correct | {total_chars - correct_chars}"
        + " incorrect"
    )

    test_results_file = open(output_directory + "/results.csv", "a")
    mistyped_words_file = open(output_directory + "/mistyped_words.csv", "a")
    char_speeds_file = open(output_directory + "/char_speeds.csv", "a")
    word_speeds_file = open(output_directory + "/word_speeds.csv", "a")

    test_results_writer = csv.writer(test_results_file, lineterminator="\n")
    row = [timestamp, wpm, accuracy, actual_duration, duration, hash]
    test_results_writer.writerow(row)

    chars, times = zip(*char_times)
    char_durations = [t1 - t0 for t0, t1 in zip(times, times[1:])]
    char_speeds_writer = csv.writer(char_speeds_file, lineterminator="\n")
    for char, duration in zip(chars, char_durations):
        char_speeds_writer.writerow([char, duration, 12 / duration, timestamp])

    word = ""
    word_i = 0
    word_duration = 0
    word_durations = []
    mistyped_writer = csv.writer(mistyped_words_file, lineterminator="\n")
    word_speeds_writer = csv.writer(word_speeds_file, lineterminator="\n")
    for char, duration in zip(chars, char_durations):
        if char.isspace():
            if word == words[word_i]:
                word_durations.append((word, word_duration))
                word_speeds_writer.writerow(
                    [
                        word,
                        word_duration,
                        len(word) * 12 / word_duration,
                        timestamp,
                    ]
                )
            else:
                mistyped_writer.writerow([words[word_i], word, timestamp])

            word_i += 1
            word = ""
            word_duration = 0
        else:
            word += char
            word_duration += duration


def draw(term, rows, words, colors, word_i, text, wpm, time_passed):
    """Text wraps the `words` list to the terminal width, and prints `rows`
    lines of wrapped words coloured with `colors` starting with the line
    containing the current word that is being typed.
    Then, if there is space, prints `prompt` + `text` + `stats`.
    """

    def join(words, length):
        eol = term.clear_eol if length + len(words) - 1 < term.width else ""
        return " ".join(line_words) + eol

    echo = partial(print, end="", flush=True, file=term.stream)
    allowed_height = min(term.height, rows)

    len_line = 0
    line_words = []
    line_height = None
    for i, (word, color) in enumerate(zip(words, colors)):
        if len_line + len(word) + len(line_words) > term.width:
            if line_height is not None:
                echo(term.move_yx(line_height, 0) + join(line_words, len_line))
                line_height += 1
                if line_height >= allowed_height:
                    break

            len_line = 0
            line_words = []

        if i == word_i:
            line_height = 0

        len_line += len(word)
        line_words.append(color + word + term.normal)

    else:
        echo(term.move_yx(line_height, 0) + join(line_words, len_line))
        line_height += 1

    if allowed_height > 1:
        prompt = ">>>"
        timestamp = strftime("%H:%M:%S", gmtime(time_passed))
        stats = f"{wpm:3d} wpm | {timestamp}"
        n = term.width - len(prompt) - len(stats)
        echo(term.move_yx(line_height, 0) + f"{prompt}{text[:n]: <{n}}{stats}")

    for i in range(1, allowed_height - line_height + 1):
        echo(term.move_yx(line_height + i, 0) + term.clear_eol)


def parse_args():
    """Parses `sys.argv` and returns a dictionary suitable for `main`."""

    def dir_path(string):
        if os.path.isdir(string):
            return string
        else:
            raise NotADirectoryError(string)

    parser = ArgumentParser(epilog=doc, formatter_class=RawTextHelpFormatter)

    default = "(default: %(default)s)"
    basedir = os.path.dirname(__file__)
    parser.add_argument(
        "-d",
        "--duration",
        type=float,
        default=None,
        help="duration in seconds "
        + "(default: <infinity> when a custom test is given, 60 seconds "
        + "otherwise)",
    )
    parser.add_argument(
        "--hash",
        type=str,
        default=None,
        help="custom hash (generated from input by default)",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=FileType("r"),
        default=sys.stdin,
        help="file to read words from (default: sys.stdin)",
    )
    parser.add_argument(
        "-o",
        "--output-directory",
        type=dir_path,
        default=basedir + "/results",
        help="file to store results in\n" + default,
    )
    parser.add_argument(
        "-s", "--shuffle", action="store_true", help="shuffle words " + default
    )
    parser.add_argument(
        "-r",
        "--rows",
        type=int,
        default=2,
        help="number of test rows to show " + default,
    )

    return dict(parser.parse_args()._get_kwargs(), help=parser.print_help)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3

import os
import re
import sys
import math
import warnings
from io import StringIO
from collections import deque, defaultdict
from argparse import ArgumentParser, RawTextHelpFormatter, FileType
from functools import partial
from itertools import cycle

from matplotlib import rcParams
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from typetest.utils import damerau_levenshtein_distance

warnings.simplefilter("ignore", np.RankWarning)
warnings.simplefilter("error", UserWarning)
rcParams.update({"figure.autolayout": True})
sns.set(font_scale=1.5)
filename = os.path.basename(sys.argv[0])
doc = f"""example:
  {filename}
  {filename} wpm
  {filename} char word
"""

known_hashes = {
    "da4846a3c2a8469dd77c921ab0b0bcd506b6e9f3": "300 most common english "
    + "words",
    "275eb003c4fba77d7e61893c3d9fa869822e06c8": "1000 most common english "
    + "words (no double letters)",
}


def run():
    """Parse command line arguments and run main"""
    main(**parse_args())


def show_diagram():
    """Do `plt.show()`, confirming a GUI matplotlib backend is available."""
    try:
        plt.show()
    except UserWarning:
        print("No matplotlib backend found that supports drawing charts.")
        print("Solve this by installing tk!")
        print("For example:")
        print("  sudo apt-get install python-tk")
        print("  sudo pacman -S tk")


def main(graphs, output, mistyped, char_speeds, word_speeds, help):
    """Draw diagrams the user has requested."""
    is_word = partial(re.match, r"^[a-z]+$")
    if "wpm" in graphs:
        plot_wpm(output)
    if "char" in graphs:
        plot_char_speeds(char_speeds, filter_func=str.islower)
    if "word" in graphs:
        plot_n_best_word_speeds(word_speeds, 50, filter_func=is_word)
    if "dist" in graphs:
        plot_word_wpm_distribution(word_speeds, filter_func=is_word)
    if "mistypes" in graphs:
        plot_mistypes_distribution(mistyped, filter_func=is_word)


def plot_wpm(output):
    """Reads `output` and plots typing speeds uniformly apart.
    Adds a trendline.
    """
    df = pd.read_csv(
        output,
        header=None,
        names=[
            "timestamp",
            "wpm",
            "accuracy",
            "actual_duration",
            "duration",
            "hash",
        ],
    )

    if len(df) < 2:
        print(
            "More data is needed, before analysing is possible. "
            + "A minimum of 2 tests is required."
        )
        return

    df.timestamp = pd.to_datetime(df.timestamp)
    # df = df.set_index(df.timestamp)

    min_wpm = None
    gdf = defaultdict(lambda: [[], pd.DataFrame()])
    for index, row in df.iterrows():
        h = row["hash"]
        indexes, hdf = gdf[h]
        indexes.append(index)
        hdf = hdf.append(row)
        gdf[h] = indexes, hdf
        if min_wpm is None or row["wpm"] < min_wpm:
            min_wpm = row["wpm"]

    # grouped = sorted(gdf.items(), key=lambda x: x[1][1]['wpm'].mean(),
    #                  reverse=True)
    grouped = gdf.items()

    fig, ax = plt.subplots()
    colors = cycle(sns.color_palette())
    for h, (indexes, hdf) in grouped:
        if h in known_hashes:
            h = known_hashes[h]
        x = indexes
        y = hdf.wpm
        color = next(colors)
        ax.plot(x, y, color=color, lw=3, label=h)
        # ax.fill_between(x, y, min_wpm, facecolor=color, label=h)
        trendline = np.poly1d(np.polyfit(x, y, 1))(x)
        ax.plot(x, trendline, "-", lw=4, color="white")
        ax.plot(x, trendline, "--", lw=2, color=color, label="trendline")

    ax.plot(df.accuracy, color="white", lw=4, alpha=0.5)
    ax.plot(
        df.accuracy,
        color=next(colors),
        lw=1.5,
        label="accuracy [%]",
        alpha=0.5,
    )

    ax.set_title("typing speed per typing test")
    ax.set_xlabel("")
    ax.set_ylabel("typing speed [wpm]")
    ax.legend()

    # ticks = plt.yticks()[0]
    # plt.yticks(np.arange(0, ticks[-1], 10))
    plt.xticks(df.index, df.timestamp.dt.date, rotation=90)
    # label only every 50th tick
    for i, label in enumerate(ax.xaxis.get_ticklabels()):
        if i % math.ceil(len(df) / 50):
            label.set_visible(False)

    show_diagram()


def plot_char_speeds(char_speeds, size=10000, filter_func=lambda c: True):
    """Reads last `size` lines of `char_speeds` and groups them by characters.
    Removes lowest and highest 10% and boxplots the data.

    filter_func: function taking a `char` returning `True` if char should be
    plotted, `False` otherwise. By default plots all characters.
    """
    q = deque(char_speeds, maxlen=size)
    df = pd.read_csv(
        StringIO("".join(q)),
        header=None,
        names=["char", "duration", "wpm", "timestamp"],
    )

    gdf = filter(
        lambda t: filter_func(t[1]["char"].iloc[0]), df.groupby(["char"])
    )
    wpms = []
    chars = []
    means = []
    for char, df in gdf:
        if filter_func(char):
            q1 = df["wpm"].quantile(0.1)  # noqa
            q3 = df["wpm"].quantile(0.9)  # noqa
            wpm = df.query("@q1 <= wpm <= @q3")["wpm"]
            chars.append(char)
            wpms.append(wpm)
            means.append(wpm.mean())

    assert chars, "Not enough data"
    fig, ax = plt.subplots()

    ax.boxplot(wpms, labels=chars)
    mean = round(sum(means) / len(means))
    ax.axhline(y=mean, color="r", linestyle="-", label=f"mean {mean} wpm")

    ax.set_title(f"typing speed per character of last {size} characters")
    ax.set_xlabel("")
    ax.set_ylabel("typing speed [wpm]")
    ax.legend()

    ticks = plt.yticks()[0]
    plt.yticks(np.arange(0, ticks[-1], 10))

    show_diagram()


def plot_n_best_word_speeds(word_speeds, n, filter_func=lambda w: True):
    """Loads all words from `word_speeds` and groups them by word."""
    half = n // 2
    df = pd.read_csv(
        word_speeds,
        header=None,
        names=["word", "duration", "wpm", "timestamp"],
    )

    gdf = list(filter(lambda t: filter_func(t[0]), df.groupby(["word"])))

    with open("least_typed_words", "w") as f:
        words_by_freq = list(zip(*sorted(gdf, key=lambda t: len(t[1]))))[0]
        f.write(" ".join(words_by_freq))

    gdf = sorted(gdf, key=lambda x: x[1]["wpm"].median())

    with open("worst_words", "w") as f:
        words_by_median = list(zip(*gdf))[0]
        f.write(" ".join(words_by_median))

    first_half = deque(maxlen=half)
    second_half = deque(maxlen=half)
    for word, df in gdf:
        if len(first_half) < first_half.maxlen:
            first_half.append((word, df["wpm"], df["wpm"].mean()))
        else:
            second_half.append((word, df["wpm"], df["wpm"].mean()))

    words, wpms, means = zip(*list(first_half) + list(second_half))
    mean = round(sum(means) / len(means))

    assert words, "Not enough data"

    fig, ax = plt.subplots()

    ax.boxplot(wpms, labels=words)
    ax.axhline(y=mean, color="r", linestyle="-", label=f"mean {mean} wpm")

    ax.set_title(f"worst and best {half} words")
    ax.set_xlabel("")
    ax.set_yscale("linear")
    ax.set_ylabel("typing speed [wpm]")
    ax.legend()

    ticks = plt.yticks()[0]
    plt.yticks(np.arange(ticks[0], ticks[-1], 10))
    plt.xticks(rotation=90)

    show_diagram()


def plot_word_wpm_distribution(word_speeds, filter_func=lambda c: True):
    """Plots a distribution over average speeds of unique words."""
    df = pd.read_csv(
        word_speeds,
        header=None,
        names=["word", "duration", "wpm", "timestamp"],
    )

    gdf = list(filter(lambda t: filter_func(t[0]), df.groupby(["word"])))
    wpms = [df["wpm"].median() for word, df in gdf]

    ax = sns.histplot(wpms, kde=True, stat="probability")
    ax.set_title("percentage of words typed at a certain speed")
    ax.set_xlabel("typing speed in wpm")
    ax.set_ylabel("percentage of words")
    show_diagram()


def plot_mistypes_distribution(mistyped, filter_func=lambda c: True):
    """Plots a pie chart representing the shares of numbers of mistakes in
    mistyped words.
    """
    df = pd.read_csv(
        mistyped, header=None, names=["word", "mistype", "timestamp"]
    )

    mistakes = defaultdict(int)
    words, mistypes = df["word"], df["mistype"]
    for word, mistype in zip(words, mistypes):
        if abs(len(word) - len(mistype)) < 10:
            distance = damerau_levenshtein_distance(word, mistype)
            if distance >= max(len(word), len(mistype)):
                # wrong word has been typed
                continue
            if distance > 3 and word.startswith(mistype):
                # word might be formed of multiple words
                # if mistype in dictionary:
                continue

            mistakes[distance] += 1

    fig, ax = plt.subplots()
    labels, sizes = zip(*sorted(mistakes.items()))
    explode = [0] + [0.2] * (len(mistakes) - 1)
    ax.pie(sizes, labels=labels, autopct="%1.1f%%", explode=explode)
    # ax = sns.histplot(mistakes, stat='probability')
    ax.set_title("number of mistakes made when typing a word")
    show_diagram()


def parse_args():
    """Parses `sys.argv` and returns a dictionary suitable for `main`."""
    parser = ArgumentParser(epilog=doc, formatter_class=RawTextHelpFormatter)

    default = "(default: %(default)s)"
    basedir = os.path.dirname(__file__)
    resultsdir = "results"
    parser.add_argument(
        "graphs",
        type=str,
        nargs="*",
        default=["wpm"],
        help="graphs to plot: wpm char word dist mistypes\n" + default,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=FileType("r"),
        default=f"{basedir}/{resultsdir}/results.csv",
        help="file to store results in\n" + default,
    )
    parser.add_argument(
        "-m",
        "--mistyped",
        type=FileType("r"),
        default=f"{basedir}/{resultsdir}/mistyped_words.csv",
        help="file to store mistyped words in\n" + default,
    )
    parser.add_argument(
        "-c",
        "--char_speeds",
        type=FileType("r"),
        default=f"{basedir}/{resultsdir}/char_speeds.csv",
        help="file to store character speeds in\n" + default,
    )
    parser.add_argument(
        "-w",
        "--word_speeds",
        type=FileType("r"),
        default=f"{basedir}/{resultsdir}/word_speeds.csv",
        help="file to store word speeds in\n" + default,
    )

    return dict(parser.parse_args()._get_kwargs(), help=parser.print_help)


if __name__ == "__main__":
    run()



import unittest

from typetest.utils import damerau_levenshtein_distance


class TestUtils(unittest.TestCase):
    def test_damerau_levenshtein_distance(self):
        self.assertEqual(damerau_levenshtein_distance("aa", "ab"), 1)
        self.assertEqual(damerau_levenshtein_distance("aa", "abc"), 2)
        self.assertEqual(damerau_levenshtein_distance("az", "abc"), 2)
        self.assertEqual(damerau_levenshtein_distance("Aa", "abc"), 3)
        self.assertEqual(damerau_levenshtein_distance("abc", "abc"), 0)
        self.assertEqual(damerau_levenshtein_distance("$@", "abc"), 3)
        self.assertEqual(damerau_levenshtein_distance("cba", "abc"), 2)
        self.assertEqual(damerau_levenshtein_distance("abacus", "abc"), 3)



"""Various utility functions."""


def damerau_levenshtein_distance(word_1: str, word_2: str) -> int:
    """Calculates the distance between two words."""
    inf = len(word_1) + len(word_2)
    table = [
        [inf for _ in range(len(word_1) + 2)] for _ in range(len(word_2) + 2)
    ]

    for i in range(1, len(word_1) + 2):
        table[1][i] = i - 1
    for i in range(1, len(word_2) + 2):
        table[i][1] = i - 1

    last_encountered_cols = {}
    for col, char_1 in enumerate(word_1, 2):
        last_row = 0
        for row, char_2 in enumerate(word_2, 2):
            last_encountered_col = last_encountered_cols.get(char_2, 0)

            addition = table[row - 1][col] + 1
            deletion = table[row][col - 1] + 1
            substitution = table[row - 1][col - 1] + (
                0 if char_1 == char_2 else 1
            )

            transposition = (
                table[last_row - 1][last_encountered_col - 1]
                + (col - last_encountered_col - 1)
                + (row - last_row - 1)
                + 1
            )

            table[row][col] = min(
                addition, deletion, substitution, transposition
            )

            if char_1 == char_2:
                last_row = row
        last_encountered_cols[char_1] = col

    return table[len(word_2) + 1][len(word_1) + 1]
